# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: MetadataSchema.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x14MetadataSchema.proto"V\n\tPointData\x12\x14\n\x08position\x18\x01 \x03(\x01\x42\x02\x10\x01\x12\x11\n\x05\x63olor\x18\x02 \x03(\rB\x02\x10\x01\x12\x10\n\x08geneName\x18\x03 \x01(\t\x12\x0e\n\x06\x63\x65llId\x18\x04 \x01(\t"B\n\x08TileData\x12\x1e\n\npointsData\x18\x01 \x03(\x0b\x32\n.PointData\x12\x16\n\x0enumberOfPoints\x18\x02 \x01(\r",\n\nColumnData\x12\x1e\n\x0b\x63olumnTiles\x18\x01 \x03(\x0b\x32\t.TileData".\n\tLevelData\x12!\n\x0clevelColumns\x18\x01 \x03(\x0b\x32\x0b.ColumnData"%\n\x08Metadata\x12\x19\n\x05level\x18\x01 \x03(\x0b\x32\n.LevelDatab\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'MetadataSchema_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
    DESCRIPTOR._loaded_options = None
    _globals['_POINTDATA'].fields_by_name['position']._loaded_options = None
    _globals['_POINTDATA'].fields_by_name['position']._serialized_options = b'\020\001'
    _globals['_POINTDATA'].fields_by_name['color']._loaded_options = None
    _globals['_POINTDATA'].fields_by_name['color']._serialized_options = b'\020\001'
    _globals['_POINTDATA']._serialized_start = 24
    _globals['_POINTDATA']._serialized_end = 110
    _globals['_TILEDATA']._serialized_start = 112
    _globals['_TILEDATA']._serialized_end = 178
    _globals['_COLUMNDATA']._serialized_start = 180
    _globals['_COLUMNDATA']._serialized_end = 224
    _globals['_LEVELDATA']._serialized_start = 226
    _globals['_LEVELDATA']._serialized_end = 272
    _globals['_METADATA']._serialized_start = 274
    _globals['_METADATA']._serialized_end = 311
# @@protoc_insertion_point(module_scope)
